# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import math
import os
import random
import sys
import warnings
import numpy as np
import torch

from typing import Dict, List, Optional, Tuple
from torch import nn
from torch import optim
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset

from towhee.cnn_trainer.trainer_callback import (
    CallbackHandler,
    DefaultFlowCallback,
    PrinterCallback,
    ProgressCallback,
    TrainerCallback,
    TrainerControl,
    TrainerState,
)
from towhee.cnn_trainer.trainer_utils import (
    PREFIX_CHECKPOINT_DIR,
    TrainOutput,
)
from towhee.cnn_trainer.training_args import TrainingArguments
from towhee.cnn_trainer.utils import logging

DEFAULT_CALLBACKS = [DefaultFlowCallback]
# DEFAULT_PRO = ProgressCallback

logger = logging.get_logger(__name__)

WEIGHTS_NAME = "pytorch_model.bin"


class PyTorchCNNTrainer:
    """
    PyTorchCNNTrainer is a simple but feature-complete training and eval loop for PyTorch.

    Args:
        model (:obj:`torch.nn.Module`):
            The model to train, evaluate or use for predictions.
        args (:class:`~towhee.TrainingArguments`):
            The arguments to tweak for training. Will default to a basic instance of
            :class:`~towhee.TrainingArguments` with the ``output_dir`` set to a directory named
            `tmp_trainer` in the current directory if not provided.
        train_dataset (:obj:`torch.utils.data.dataset.Dataset`):
            The dataset to use for training.
        eval_dataset (:obj:`torch.utils.data.dataset.Dataset`, `optional`):
             The dataset to use for evaluation.
        callbacks (List of :obj:`~towhee.TrainerCallback`, `optional`):
            A list of callbacks to customize the training loop.

    """

    def __init__(
            self,
            model: nn.Module = None,
            args: TrainingArguments = None,
            train_dataset: Optional[Dataset] = None,
            eval_dataset: Optional[Dataset] = None,
            callbacks: Optional[List[TrainerCallback]] = None,
            optimizers: Tuple[torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR]
            = (None, None),
    ):
        if args is None:
            output_dir = "tmp_trainer"
            logger.info("No `TrainingArguments` passed, using `output_dir.")
            args = TrainingArguments(output_dir=output_dir)
        self.args = args

        if model is None:
            raise RuntimeError("`Trainer` requires either a `model` or `model_init` argument")

        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset

        self.model = model

        self.optimizer, self.lr_scheduler = optimizers
        default_callbacks = DEFAULT_CALLBACKS
        callbacks = default_callbacks if callbacks is None else default_callbacks + callbacks
        self.callback_handler = CallbackHandler(
            callbacks, self.model, self.optimizer, self.lr_scheduler
        )
        self.add_callback(PrinterCallback if self.args.disable_tqdm else ProgressCallback)

        os.makedirs(self.args.output_dir, exist_ok=True)

        if args.max_steps > 0:
            logger.info("max_steps is given.")

        if train_dataset is not None and not isinstance(train_dataset, collections.abc.Sized) and args.max_steps <= 0:
            raise ValueError("train_dataset does not implement __len__, max_steps has to be specified")

        self.state = TrainerState()
        # control the save condition
        self.control = TrainerControl()
        # Internal variable to count flos in each process, will be accumulated in `self.state.total_flos` then
        # returned to 0 every time flos need to be logged
        self.current_flcurrent_flosos = 0
        default_label_names = (
            ["labels"]
        )
        self.label_names = default_label_names if self.args.label_names is None else self.args.label_names
        self.control = self.callback_handler.on_init_end(self.args, self.state, self.control)

    def add_callback(self, callback):
        """
        Add a callback to the current list of :class:`~towhee.TrainerCallback`.

        Args:
           callback (:obj:`type` or :class:`~towhee.TrainerCallback`):
               A :class:`~towhee.TrainerCallback` class or an instance of a :class:`~towhee.TrainerCallback`.
               In the first case, will instantiate a member of that class.
        """
        self.callback_handler.add_callback(callback)

    def get_train_dataloader(self) -> DataLoader:
        """
        Returns the training :class:`~torch.utils.data.DataLoader`.
        """
        if self.train_dataset is None:
            raise ValueError("Trainer: training requires a train_dataset.")

        train_dataset = self.train_dataset

        return DataLoader(
            train_dataset,
            batch_size=self.args.train_batch_size,
            shuffle=True
        )

    def create_optimizer_and_scheduler(self):
        """
        Setup the optimizer and the learning rate scheduler.
        """
        self.create_optimizer()
        self.create_scheduler()

    def create_optimizer(self):
        """
        Setup the optimizer.
        """
        self.optimizer = optim.SGD(self.model.parameters(), lr=0.01, momentum=0.9)

    def create_scheduler(self):
        """
        Setup the scheduler. The optimizer of the cnn_trainer must have been set up before this method is called.
        """
        self.lr_scheduler = optim.lr_scheduler.StepLR(self.optimizer, step_size=3, gamma=0.1)

    def num_examples(self, dataloader: DataLoader) -> int:
        """
        Helper to get number of samples in a :class:`~torch.utils.data.DataLoader` by accessing its dataset.

        Will raise an exception if the underlying dataset does not implement method :obj:`__len__`
        """
        return len(dataloader.dataset)

    def train(self):
        """
        Main training entry point.
        """
        args = self.args

        # Keeping track whether we can can len() on the dataset or not
        train_dataset_is_sized = isinstance(self.train_dataset, collections.abc.Sized)

        # Data loader and number of training steps
        train_dataloader = self.get_train_dataloader()

        # Setting up training control variables:
        # number of training epochs: num_train_epochs
        # number of training steps per epoch: num_update_steps_per_epoch
        # total number of training steps to execute: max_steps
        total_train_batch_size = args.train_batch_size
        if train_dataset_is_sized:
            num_update_steps_per_epoch = len(train_dataloader)
            num_update_steps_per_epoch = max(num_update_steps_per_epoch, 1)
            if args.max_steps > 0:
                max_steps = args.max_steps
                num_train_epochs = args.max_steps // num_update_steps_per_epoch + int(
                    args.max_steps % num_update_steps_per_epoch > 0
                )
            else:
                max_steps = math.ceil(args.num_train_epochs * num_update_steps_per_epoch)
                num_train_epochs = math.ceil(args.num_train_epochs)
        else:
            max_steps = args.max_steps
            # Setting a very large number of epochs so we go as many times as necessary over the iterator.
            num_train_epochs = sys.maxsize

        self.create_optimizer_and_scheduler()

        self.state = TrainerState()

        model = self.model

        # Train!
        num_examples = (
            self.num_examples(train_dataloader) if train_dataset_is_sized else total_train_batch_size * args.max_steps
        )

        logger.info("***** Running training *****")
        logger.info("  Num examples = %d", num_examples)
        logger.info("  Num Epochs = %d", num_train_epochs)
        logger.info("  Total train batch size (w. parallel, distributed & accumulation) = %d", total_train_batch_size)
        logger.info("  Total optimization steps = %d", max_steps)

        self.state.epoch = 0
        epochs_trained = 0

        # Update the references
        self.callback_handler.model = self.model
        self.callback_handler.optimizer = self.optimizer
        self.callback_handler.lr_scheduler = self.lr_scheduler
        self.callback_handler.train_dataloader = train_dataloader
        # This should be the same if the state has been saved but in case the training arguments changed, it's safer
        # to set this after the load.
        self.state.max_steps = max_steps
        self.state.num_train_epochs = num_train_epochs

        tr_loss = torch.tensor(0.0)
        tr_corrects = 0
        # _total_loss_scalar is updated everytime .item() has to be called on tr_loss and stores the sum of all losses
        self._total_loss_scalar = 0.0
        # model.zero_grad()

        self.control = self.callback_handler.on_train_begin(args, self.state, self.control)

        for epoch in range(epochs_trained, num_train_epochs):
            epoch_iterator = train_dataloader

            steps_in_epoch = (
                len(epoch_iterator) if train_dataset_is_sized else args.max_steps
            )
            self.control = self.callback_handler.on_epoch_begin(args, self.state, self.control)

            for step, inputs in enumerate(epoch_iterator):
                loss, corrects = self.training_step(model, inputs)
                tr_loss += loss
                tr_corrects += corrects

                # Optimizer step
                optimizer_was_run = True
                self.optimizer.step()

                if optimizer_was_run:
                    self.lr_scheduler.step()

                self.optimizer.zero_grad()
                self.state.global_step += 1
                self.state.epoch = epoch + (step + 1) / steps_in_epoch
                self.control = self.callback_handler.on_step_end(args, self.state, self.control)

            self.control = self.callback_handler.on_epoch_end(args, self.state, self.control)
            self._maybe_log_save_evaluate(tr_loss, tr_corrects, num_examples)

            if self.control.should_training_stop:
                break

        logger.info("\n\nTraining completed. Do not forget to share your model on huggingface.co/models =)\n\n")

        # add remaining tr_loss
        self._total_loss_scalar += tr_loss.item() / num_examples
        # train_loss = self._total_loss_scalar / self.state.global_step
        train_loss = self._total_loss_scalar / num_examples

        self.control = self.callback_handler.on_train_end(args, self.state, self.control)

        return TrainOutput(self.state.global_step, train_loss)

    def _maybe_log_save_evaluate(self, tr_loss, tr_corrects, num_examples):
        if self.control.should_log:
            logs: Dict[str, float] = {}
            tr_loss_scalar = tr_loss.item()
            epoch_loss = tr_loss_scalar / num_examples
            epoch_acc = tr_corrects / num_examples

            # reset tr_loss to zero
            tr_loss -= tr_loss
            tr_corrects -= tr_corrects

            logs["loss"] = epoch_loss
            logs["accuracy"] = epoch_acc

            self._total_loss_scalar += tr_loss_scalar
            self.log(logs)

        self.control.should_save = False
        if self.control.should_save:
            self._save_checkpoint()
            self.control = self.callback_handler.on_save(self.args, self.state, self.control)

    def _save_checkpoint(self):
        # Save model checkpoint
        checkpoint_folder = f"{PREFIX_CHECKPOINT_DIR}-{self.state.global_step}"

        run_dir = self.args.output_dir

        output_dir = os.path.join(run_dir, checkpoint_folder)
        self.save_model(output_dir)

        # Save optimizer and scheduler
        if self.args.should_save:
            torch.save(self.optimizer.state_dict(), os.path.join(output_dir, "optimizer.pt"))
            with warnings.catch_warnings(record=True):
                torch.save(self.lr_scheduler.state_dict(), os.path.join(output_dir, "scheduler.pt"))

        # Save the Trainer state
        if self.args.should_save:
            self.state.save_to_json(os.path.join(output_dir, "trainer_state.json"))

        # Save RNG state in non-distributed training
        rng_states = {
            "python": random.getstate(),
            "numpy": np.random.get_state(),
            "cpu": torch.random.get_rng_state(),
        }
        if torch.cuda.is_available():
            rng_states["cuda"] = torch.cuda.random.get_rng_state()

    def log(self, logs: Dict[str, float]) -> None:
        """
        Log :obj:`logs` on the various objects watching training.

        Args:
            logs (:obj:`Dict[str, float]`):
                The values to log.
        """
        if self.state.epoch is not None:
            logs["epoch"] = round(self.state.epoch, 2)

        output = {**logs, **{"step": self.state.global_step}}
        self.state.log_history.append(output)
        self.control = self.callback_handler.on_log(self.args, self.state, self.control, logs)

    def training_step(self, model: nn.Module, inputs: List[torch.Tensor]) -> torch.Tensor:
        """
        Perform a training step on a batch of inputs.

        Args:
            model (:obj:`nn.Module`):
                The model to train.
            inputs (:obj:`Dict[str, Union[torch.Tensor, Any]]`):
                The inputs and targets of the model.

        Return:
            :obj:`torch.Tensor`: The tensor with training loss on this batch.
        """
        model.train()
        # inputs = self._prepare_inputs(inputs)

        loss, corrects = self.compute_loss(model, inputs)

        if self.args.n_gpu > 1:
            loss = loss.mean()  # mean() to average on multi-gpu parallel training

        loss.backward()

        return loss.detach(), corrects

    def compute_loss(self, model, inputs):
        """
        How the loss is computed by PyTorchCNNTrainer. By default, all models return the loss in the first element.

        Subclass and override for custom behavior.
        """
        labels = inputs[1]
        with torch.set_grad_enabled(True):
            outputs = model(inputs[0])
            _, preds = torch.max(outputs, 1)

        if labels is not None:
            criterion = nn.CrossEntropyLoss()
            loss = criterion(outputs, labels)
            corrects = torch.sum(preds == labels.data)

        # return (loss, outputs) if return_outputs else loss
        return loss, corrects

    def save_model(self, output_dir: Optional[str] = None):
        """
        Will save the model.
        """

        if output_dir is None:
            output_dir = self.args.output_dir

        state_dict = self.model.state_dict()
        if self.args.should_save:
            self._save(output_dir, state_dict)

    def _save(self, output_dir: Optional[str] = None, state_dict=None):
        # If we are executing this function, we are the process zero, so we don't check for that.
        output_dir = output_dir if output_dir is not None else self.args.output_dir
        os.makedirs(output_dir, exist_ok=True)
        logger.info("Saving model checkpoint to %s", output_dir)
        torch.save(state_dict, os.path.join(output_dir, WEIGHTS_NAME))

        # Good practice: save your training arguments together with the trained model
        torch.save(self.args, os.path.join(output_dir, "training_args.bin"))



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT_ WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
import yaml
import copy
from unittest import mock
from collections import namedtuple

from towhee.dag.base_repr import BaseRepr
from towhee.dag.graph_repr import GraphRepr

from towhee.tests.test_util import GRAPH_TEST_YAML, GRAPH_TEST_ISO_DF_YAML, GRAPH_TEST_ISO_OP_YAML, GRAPH_TEST_LOOP_YAML


@unittest.skip('')
class TestRepr(unittest.TestCase):
    """Basic test cases for `hRepr`.
    """

    def test_base_repr(self):
        self.assertTrue(BaseRepr.is_valid({'test': 1}, {'test'}))
        self.assertFalse(BaseRepr.is_valid({'test': 1}, {'not_exist'}))

    def _check_graph(self, g_repr):
        # test dataframes generation
        self.assertEqual(len(g_repr.dataframes.values()), 3)
        self.assertTrue('test_df_1' in g_repr.dataframes)
        self.assertTrue('test_df_2' in g_repr.dataframes)
        self.assertEqual(
            g_repr.dataframes['test_df_1'].columns[0].vtype, 'int')
        # test operators generation
        self.assertEqual(len(g_repr.operators.values()), 2)
        self.assertTrue('test_op_1' in g_repr.operators)
        self.assertTrue('test_op_2' in g_repr.operators)
        # check details in operators
        self.assertEqual(g_repr.operators['test_op_1'].inputs, [
                         {'df': 'test_df_1', 'name': 'k1', 'col': 0}])
        self.assertEqual(g_repr.operators['test_op_2'].inputs, [
                         {'df': 'test_df_2', 'name': 'k1', 'col': 0}])
        self.assertEqual(g_repr.operators['test_op_2'].outputs, [
                         {'df': 'test_df_3'}])
        self.assertEqual(g_repr.operators['test_op_2'].init_args, {
                         'arg1': 1, 'arg2': 'test'})
        self.assertEqual(
            g_repr.operators['test_op_2'].iter_info, {'type': 'map'})

    def test_graph_repr(self):
        # Check if the inforamtion is loaded properly
        # Load from YAML file
        self._check_graph(GraphRepr.from_yaml(GRAPH_TEST_YAML))
        # Load from string
        with open(GRAPH_TEST_YAML, 'r', encoding='utf-8') as f:
            self._check_graph(GraphRepr.from_yaml(f.read()))
        # Load from url
        with mock.patch('requests.get') as mock_get:
            with open(GRAPH_TEST_YAML, 'r', encoding='utf-8') as f:
                MockResponse = namedtuple('MockResponse', ['text'])
                mock_get.return_value = MockResponse(f.read())
                self._check_graph(GraphRepr.from_yaml('http://mock.yaml'))

    def test_isolation_or_loop(self):
        # Raise error if given false source information that contain loop or isolation
        self.assertRaises(ValueError, GraphRepr.from_yaml,
                          GRAPH_TEST_ISO_DF_YAML)
        self.assertRaises(ValueError, GraphRepr.from_yaml,
                          GRAPH_TEST_ISO_OP_YAML)
        self.assertRaises(ValueError, GraphRepr.from_yaml,
                          GRAPH_TEST_LOOP_YAML)

        # A proper graph does not contain isolation and loopa
        with open(GRAPH_TEST_YAML, 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f.read())
        graph = GraphRepr.from_dict(data)
        non_iso_df = graph.get_isolated_df()
        self.assertFalse(bool(non_iso_df))
        non_iso_op = graph.get_isolated_op()
        self.assertFalse(bool(non_iso_op))
        none_loop = graph.get_loop()
        self.assertFalse(bool(none_loop))
        # If the graph contains isolated dataframes
        with open(GRAPH_TEST_ISO_DF_YAML, 'r', encoding='utf-8') as f:
            iso_data = yaml.safe_load(f.read())
        iso_df_graph = GraphRepr.from_dict(iso_data)
        iso_df = iso_df_graph.get_isolated_df()
        self.assertTrue(bool(iso_df))
        # If the graph contains isolated operators
        with open(GRAPH_TEST_ISO_OP_YAML, 'r', encoding='utf-8') as f:
            iso_data = yaml.safe_load(f.read())
        iso_op_graph = GraphRepr.from_dict(iso_data)
        iso_op = iso_op_graph.get_isolated_op()
        self.assertTrue(bool(iso_op))
        # If the graph contains loops
        with open(GRAPH_TEST_LOOP_YAML, 'r', encoding='utf-8') as f:
            iso_data = yaml.safe_load(f.read())
        loop_graph = GraphRepr.from_dict(iso_data)
        loop = loop_graph.get_loop()
        self.assertTrue(bool(loop))

    def test_error_yaml(self):
        with open(GRAPH_TEST_YAML, 'r', encoding='utf-8') as f:
            data = yaml.safe_load(f.read())
        err1 = copy.deepcopy(data)
        del err1['dataframes']
        with self.assertRaises(ValueError):
            GraphRepr.from_dict(err1)

        err2 = copy.deepcopy(data)
        del err2['operators'][0]['init_args']
        with self.assertRaises(ValueError):
            GraphRepr.from_dict(err2)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import json
from PIL import Image
from torchvision import transforms
import os

from towhee.trainer.models.vit.vit import ViT
from towhee.tests.mock_operators import PYTORCH_TRANSFORMER_OPERATOR_PATH, load_local_operator


class TransformerOperatorTest(unittest.TestCase):
    name = 'B_16_imagenet1k'
    pwd = os.getcwd()
    print('pwd is ' + str(pwd))
    test_dir = pwd+'/towhee/tests/models/vit/'
    #weights_path = test_dir + 'B_16_imagenet1k.pth'
    weights_path = None
    model = ViT(name, weights_path=weights_path, pretrained=True)
    img = Image.open(test_dir + 'img.jpg')
    tfms = transforms.Compose([transforms.Resize(model.image_size), transforms.ToTensor(),
                               transforms.Normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5]), ])
    img = tfms(img).unsqueeze(0)
    with open(test_dir + 'labels_map.txt', encoding='utf-8') as handler:
        labels_map = json.load(handler)
    _labels_map = []
    for i in range(1000):
        _labels_map.append(labels_map[str(i)])
    labels_map = _labels_map
    args = {'topk': 1, 'labels_map': labels_map}

    def test_transformer_operator(self):
        trans = load_local_operator(
            'pytorch_transformer_operator', PYTORCH_TRANSFORMER_OPERATOR_PATH)
        op = trans.PytorchTransformerOperator(self.model, self.args)
        outputs = op(self.img)
        print(type(outputs.predict))
        print('outputs.predict is '+outputs.predict)
        self.assertEqual(
            'giant panda, panda, panda bear, coon bear, Ailuropoda melanoleuca', outputs.predict)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
import unittest

from towhee.engine.task_executor import TaskExecutor
from towhee.engine.task import Task


class TestTaskExecutor(unittest.TestCase):
    """Basic test case for `TaskExecutor`.
    """

    def setUp(self):
        cache_path = Path(__file__).parent.parent.resolve()
        self._task_exec = TaskExecutor('', cache_path=cache_path)
        self._task_exec.start()

    def tearDown(self):
        self._task_exec.stop()

    def test_add_task_execution(self):

        # Add callback function upon completion.
        def _add_task_finish_callback(task):
            self.assertEqual(task.outputs.sum, task.inputs['num'])

        # Create a couple of tasks to execute through the executor.
        tasks = []
        hub_op_id = 'mock_operators/add_operator'
        args = {'factor': 0}
        tasks.append(Task('test', hub_op_id, args, {'num': 0}, 0))
        tasks.append(Task('test', hub_op_id, args, {'num': 1}, 1))
        tasks.append(Task('test', hub_op_id, args, {'num': 10}, 10))

        # Add finish callbacks and submit the tasks to the executor.
        for task in tasks:
            task.add_finish_handler(_add_task_finish_callback)
            self._task_exec.push_task(task)

    def test_sub_task_execution(self):

        # Add callback function upon completion.
        def _add_task_finish_callback(task):
            diff = task.inputs['a'] - task.inputs['b']
            self.assertEqual(task.outputs.diff, diff)

        # Create a couple of tasks to execute through the executor.
        tasks = []
        hub_op_id = 'mock_operators/sub_operator'
        tasks.append(Task('test', hub_op_id, {}, {'a': 0, 'b': 0}, 0))
        tasks.append(Task('test', hub_op_id, {}, {'a': 10, 'b': 20}, 1))
        tasks.append(Task('test', hub_op_id, {}, {'a': 23, 'b': -1}, 24))

        # Add finish callbacks and submit the tasks to the executor.
        for task in tasks:
            task.add_finish_handler(_add_task_finish_callback)
            self._task_exec.push_task(task)


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utilities for the Trainer.
"""

import random
import re
from typing import NamedTuple

import numpy as np
import torch


def set_seed(seed: int):
    """
    Helper function for reproducible behavior to set the seed in ``random``, ``numpy``, ``torch``.

    Args:
        seed (:obj:`int`): The seed to set.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


class TrainOutput(NamedTuple):
    global_step: int
    training_loss: float


PREFIX_CHECKPOINT_DIR = "checkpoint"
_re_checkpoint = re.compile(r"^" + PREFIX_CHECKPOINT_DIR + r"\-(\d+)$")



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod

from towhee.dataframe import DataFrame, Variable, DataFrameIterator
from typing import Dict, Optional, Tuple, Union, List, NamedTuple


class DataFrameWriter:
    '''
    Df writer
    '''

    def __init__(self, output_df: DataFrame):
        self._output_df = output_df

    def write(self, output_data: NamedTuple) -> bool:

        # TODO (jiangjunjie) deal with task exception
        if output_data is not None:
            return self._output_df.put_dict(output_data._asdict())

    def close(self):
        self._output_df.seal()


class DataFrameReader(ABC):
    """
    Read data from input dataframes, unpack and combine data.
    One op_ctx has one dataframe reader.
    """

    def __init__(self, it: DataFrameIterator, op_inputs_index: Dict[str, int]):
        self._op_inputs_index = op_inputs_index
        self._iter = it

    @abstractmethod
    def read(self) -> Union[Dict[str, any], List[Dict[str, any]]]:
        pass

    @property
    def size(self) -> int:
        return self._iter.accessible_size

    def _to_op_inputs(self, cols: Tuple[Variable]) -> Dict[str, any]:
        """
        Read from cols, combine op inputs
        """
        ret = {}
        for key, index in self._op_inputs_index.items():
            ret[key] = cols[index].value
        return ret


class MapDataFrameReader(DataFrameReader):
    """
    Map dataframe reader
    """

    def __init__(
        self,
        input_df: DataFrame,
        op_inputs_index: Dict[str, int]
    ):
        super().__init__(input_df.map_iter(), op_inputs_index)

    def read(self) -> Optional[Dict[str, any]]:
        """
        Read data from dataframe, get cols by operator_repr info
        """
        try:
            data = next(self._iter)
            if len(data) == 0:
                return {}
            return self._to_op_inputs(data[0])

        except StopIteration:
            return None


def create_reader(
    inputs: List[DataFrame],
    iter_type: str,
    inputs_index: Dict[str, int]
) -> DataFrameReader:

    if iter_type.lower() == "map":
        assert len(inputs) == 1, "%s iter takes one dataframe, but %s dataframes are found" % (
            iter_type, len(inputs)
        )
        return MapDataFrameReader(inputs[0], inputs_index)
    else:
        raise NameError("Can not find %s iters" % iter_type)


def create_writer(outputs: List[DataFrame]) -> DataFrameWriter:
    assert len(outputs) == 1
    return DataFrameWriter(outputs[0])



# Copyright 2021 Facebook and Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
from typing import Callable, Iterable, Tuple
import torch
from torch import nn
from torch.optim import Optimizer


class AdamW(Optimizer):
    """
    Implements Adam algorithm with weight decay fix as introduced in `Decoupled Weight Decay Regularization
    <https://arxiv.org/abs/1711.05101>`.
    Parameters:
        params (Iterable[nn.parameter.Parameter]):
            Iterable of parameters to optimize or dictionaries defining parameter groups.
        lr (float, optional):
            The learning rate to use.
        betas (Tuple[float,float], optional):
            Adam's betas parameters (b1, b2).
        eps (float, optional):
            Adam's epsilon for numerical stability.
        weight_decay (float, optional):
            Decoupled weight decay to apply.
        correct_bias (bool, optional):
            Whether or not to correct bias in Adam.
    """

    def __init__(
        self,
        params: Iterable[nn.parameter.Parameter],
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-6,
        weight_decay: float = 0.0,
        correct_bias: bool = True,
    ):
        if lr < 0.0:
            raise ValueError(f"Invalid learning rate: {lr} - should be >= 0.0")
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError(f"Invalid beta parameter: {betas[0]} - should be in [0.0, 1.0[")
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError(f"Invalid beta parameter: {betas[1]} - should be in [0.0, 1.0[")
        if 0.0 > eps:
            raise ValueError(f"Invalid epsilon value: {eps} - should be >= 0.0")
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, correct_bias=correct_bias)
        super().__init__(params, defaults)

    def step(self, closure: Callable = None):
        """
        Performs a single optimization step.
        Arguments:
            closure(Callable, optional):
                A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Adam does not support sparse gradients, please consider SparseAdam instead")

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # Decay the first and second moment running average coefficient
                # In-place operations to update the averages at the same time
                exp_avg.mul_(beta1).add_(grad, alpha=(1.0 - beta1))
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1.0 - beta2)
                denom = exp_avg_sq.sqrt().add_(group["eps"])

                step_size = group["lr"]
                if group["correct_bias"]:  # No bias correction for Bert
                    bias_correction1 = 1.0 - beta1 ** state["step"]
                    bias_correction2 = 1.0 - beta2 ** state["step"]
                    step_size = step_size * math.sqrt(bias_correction2) / bias_correction1

                p.data.addcdiv_(exp_avg, denom, value=-step_size)

                # Just adding the square of the weights to the loss function is *not*
                # the correct way of using L2 regularization/weight decay with Adam,
                # since that will interact with the m and v parameters in strange ways.
                #
                # Instead we want to decay the weights in a manner that doesn't interact
                # with the m/v parameters. This is equivalent to adding the square
                # of the weights to the loss with plain (non-momentum) SGD.
                # Add weight decay at the end (fixed version)
                if group["weight_decay"] > 0.0:
                    p.data.add_(p.data, alpha=(-group["lr"] * group["weight_decay"]))

        return loss



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from pathlib import Path

import torchvision
from torch import nn
from torchvision import transforms

from towhee.cnn_trainer.cnn_trainer import PyTorchCNNTrainer
from towhee.cnn_trainer.training_args import TrainingArguments
from towhee.dataset.image_datasets import PyTorchImageDataset

cache_path = Path(__file__).parent.parent.resolve()
image_path = cache_path.joinpath('dataset/kaggle_dataset_small/train')
label_file =cache_path.joinpath('dataset/kaggle_dataset_small/train/train_labels.csv')


class TrainerTest(unittest.TestCase):
    def setUp(self) -> None:
        data_transforms = {
            'train': transforms.Compose([
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
            ])
        }
        self.train_data = PyTorchImageDataset(image_path, label_file, data_transforms['train'])
        self.model = torchvision.models.resnet50(pretrained=True)
        num_ftrs = self.model.fc.in_features
        self.model.fc = nn.Linear(num_ftrs, self.train_data.num_classes)
        self.training_args = TrainingArguments(
            output_dir='./ResNet50',
            overwrite_output_dir=True,
            num_train_epochs=5,
            per_gpu_train_batch_size=4,
            prediction_loss_only=True,
        )
        self.trainer = PyTorchCNNTrainer(
            model=self.model,
            args=self.training_args,
            train_dataset=self.train_data
        )

    def test_overfit_on_small_batches(self) -> None:
        training_output = self.trainer.train()
        self.assertGreaterEqual(3.0, training_output.training_loss)


if __name__ == '__main__':
    unittest.main(verbosity=1)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import math
from urllib import request
from towhee.trainer.models.vit.vit_pretrained import PRETRAINED_MODELS


class VitPretrainedTest(unittest.TestCase):
    name = 'B_16_imagenet1k'

    def test_pretrained(self):
        assert self.name in PRETRAINED_MODELS.keys(), \
            'name should be in: ' + ', '.join(PRETRAINED_MODELS.keys())
        with request.urlopen(PRETRAINED_MODELS[self.name]['url']) as file:
            self.assertEqual(math.floor(file.status/100), 2)
            self.assertEqual(file.reason, 'OK')


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# from operator_repr import OperatorRepr
from utils import at_compile_phase
from graph_builder import get_graph_builder


class OperatorMetaclass(type):
    '''
    Metaclass for creating Operators.
    '''

    def __call__(cls, *args, **kwargs):
        """
        Return: in compile-phase, an OperatorRepr will be returned, rather than an
            Operator.
        """
        op = super().__call__(*args, **kwargs)

        if at_compile_phase():
            graph_builder = get_graph_builder()
            op_repr = graph_builder.add_op(op)
            return op_repr
        else:
            return op

    def _op_callable_check(cls, op_dict):
        """
        Test whether an Operator can be called.

        Args:

        Returns:
            True if the Operator follows all the callable rules, False other wise.

        Raises:

        Examples:
            In Towhee, each Operator should be a callable. Each interface Operator will
            define an abstract method specifying its functionality. For example,

            class Img2VecOp(Operator):
                @op_action
                @abc.abstractmethod
                def forward(self, img:Image) -> FloatVector:
                    pass

                def __call__(self, img:Image):
                    return forward(img)

            The forward function plays as the core functionality of Img2VecOp, and it
            will be called by __call__. Note that the function is decorated by
            @op_action. _op_callable_check will go through Img2VecOp's __dict__, and
            find out the methods with @op_action decoration. For each decorated
            function, OpMetaclass will check and makesure that it is overwritten by
            the subclass with identical input/output declarations.

            Here is an example class inherited Img2VecOp.

            class VGGAnimalImg2VecOp(Img2VecOp):
                def forward(self, img:Image) -> FloatVector:
                    # do the real forward works ...

            The strongly typed override mechanism guarantees that the inputs and outputs
            standard of a interface Operator is strictly followed by its subclasses.
            If not, the OpMetaclass will find such inconsistency during the subclass's
            creation time.

            An Operator may have multiple outputs. For this case, the interface
            Operator should define the outputs as a NamedTuple. For example,

            class ObjectDetectionOp(Operator):
                @op_action
                @abc.abstractmethod
                def inference(self, img:Image) -> NamedTuple('Outputs', [('bboxes', list[BBox]), ('labels', list[str]), ('scores', list[float])]):
                    pass

                def __call__(self, img:Image):
                    return inference(img)
        """
        raise NotImplementedError



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
import unittest


from PIL import Image

from towhee import pipeline, _get_pipeline_cache, _PIPELINE_CACHE_ENV
from towhee.engine.engine import EngineConfig


CACHE_PATH = Path(__file__).parent.parent.resolve()


class TestPipeline(unittest.TestCase):
    """Tests `pipeline` functionality.
    """

    def setUp(self):
        conf = EngineConfig()
        conf.cache_path = CACHE_PATH
        conf.sched_interval_ms = 20

    def test_simple_pipeline(self):
        p = pipeline('test_util/simple_pipeline', cache=str(CACHE_PATH))
        res = p(0)
        self.assertEqual(res[0], 3)

    def test_embedding_pipeline(self):
        p = pipeline('test_util/resnet50_embedding',
                     cache=str(CACHE_PATH))
        img_path = CACHE_PATH / 'dataset' / 'kaggle_dataset_small' / \
            'train' / '0021f9ceb3235effd7fcde7f7538ed62.jpg'
        img = Image.open(str(img_path))
        res = p(img)
        print('successfully got an embedding of size {0}'.format(res[0].size))
        self.assertEqual(res[0].size, 1000)

        with self.assertRaises(IndexError):
            print(p('a'))


class TestPipelineCache(unittest.TestCase):
    def test_pipeline_cache(self):
        self.assertEqual(_get_pipeline_cache(
            None), Path.home() / '.towhee/pipelines')

        os.environ[_PIPELINE_CACHE_ENV] = '/opt/.pipeline'
        self.assertEqual(_get_pipeline_cache(
            None), Path('/opt/.pipeline'))

        self.assertEqual(_get_pipeline_cache(
            '/home/mycache'), Path('/home/mycache'))


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import numpy as np

from towhee.array import (
    Array,
    full,
)


class TestArray(unittest.TestCase):
    """Basic test case for `Array`.
    """

    def test_construction_from_list(self):

        def _basic_asserts(array, data):
            self.assertEqual(array.size, len(data))

            for a, b in zip(array, data):
                if isinstance(a, np.ndarray):
                    self.assertFalse((a - b).all())
                else:
                    self.assertEqual(a, b)

            for i, v in enumerate(data):
                if isinstance(array[i], np.ndarray):
                    self.assertFalse((array[i] - v).all())
                else:
                    self.assertEqual(v, array[i])

        data = [1, 2, 3]
        array = Array(data=data)
        _basic_asserts(array, data)

        data = ["a", "b", "c"]
        array = Array(data=data)
        _basic_asserts(array, data)

        data = [3.14]
        array = Array(data=data)
        _basic_asserts(array, data)

        data = [[1, 2], [3, 4]]
        array = Array(data=data)
        _basic_asserts(array, data)

        data = []
        array = Array(data=data)
        _basic_asserts(array, data)

    def test_construction_from_scalar(self):

        def _basic_asserts(array, data):
            self.assertEqual(array.size, 1)
            self.assertEqual(array[0], data)

        data = 1
        array = Array(data=data)
        _basic_asserts(array, data)

        data = "abc"
        array = Array(data=data)
        _basic_asserts(array, data)

        data = 3.14
        array = Array(data=data)
        _basic_asserts(array, data)

    def test_construction_from_ndarray(self):

        def _basic_asserts(array, data):
            if data.size == 1:
                data = data.reshape((1,))

            self.assertEqual(array.size, data.shape[0])

            for a, b in zip(array, data):
                if isinstance(a, np.ndarray):
                    self.assertFalse((a - b).all())
                else:
                    self.assertEqual(a, b)

            for i, v in enumerate(data):
                if isinstance(array[i], np.ndarray):
                    self.assertFalse((array[i] - v).all())
                else:
                    self.assertEqual(v, array[i])

        data = np.ndarray([1, 2, 3])
        array = Array(data=data)
        _basic_asserts(array, data)

        data = np.asarray(None)
        array = Array(data=data)
        _basic_asserts(array, data)

        data = np.array([None])
        array = Array(data=data)
        _basic_asserts(array, data)

        data = np.zeros((2, 3))
        array = Array(data=data)
        _basic_asserts(array, data)

    def test_full(self):

        def _basic_asserts(array, value, n):
            self.assertEqual(array.size, n)

            for i, v in enumerate(array):
                self.assertEqual(v, value)
                self.assertEqual(array[i], value)

        n = 8
        value = 1
        array = full(shape=n, fill_value=value)
        _basic_asserts(array, value, n)

        n = 1
        value = 1
        array = full(shape=n, fill_value=value)
        _basic_asserts(array, value, n)

        n = 0
        value = 1
        array = full(shape=n, fill_value=value)
        _basic_asserts(array, value, n)

        n = 8
        value = "abc"
        array = full(shape=n, fill_value=value)
        _basic_asserts(array, value, n)

        n = 8
        value = None
        array = full(shape=n, fill_value=value)
        _basic_asserts(array, value, n)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import yaml
import requests
import os
import logging
from typing import List, Dict, Set


class BaseRepr:
    """Base representation from which all other representation objects inherit.
    Primarily implements automatic serialization into YAML/YAML-like string formats,
    along with defining other universally used properties.

    Args:
        name:
            Name of the internal object described by this representation.
    """
    def __init__(self, name: str):
        self._name = name

    @property
    def name(self):
        return self._name

    @staticmethod
    def is_valid(info: Dict, essentials: Set[str]) -> bool:
        """Check if the src is a valid YAML file to describe a representation in Towhee.

        Args:
            info(`dict`):
                The dict loaded from the source file.
        """
        info_keys = set(info.keys())
        if not isinstance(info, dict) or not essentials.issubset(info_keys):
            logging.error('Info [%s] is not valid, lack attr [%s]', str(info), essentials - info_keys)
            return False
        return True

    @staticmethod
    def load_str(string: str) -> Dict[str, any]:
        """Load the representation(s) information from a YAML file (pre-loaded as string).

        Args:
            string(`str`):
                The string pre-loaded from a YAML.

        Returns:
            The dict loaded from the YAML file that contains the representation information.
        """
        return yaml.safe_load(string)

    @staticmethod
    def load_file(file: str) -> List[dict]:
        """Load the representation(s) information from a local YAML file.

        Args:
            file(`str`):
                The file path.

        Returns:
            The list loaded from the YAML file that contains the representation information.
        """
        with open(file, 'r', encoding='utf-8') as f:
            return BaseRepr.load_str(f)

    @staticmethod
    def load_url(url: str) -> List[dict]:
        """Load the representation information from a remote YAML file.

        Args:
            url(`str`):
                The url points to the remote YAML file.

        Returns:
            The list loaded from the YAML file that contains the representation information.
        """
        src = requests.get(url, timeout=5).text
        return BaseRepr.load_str(src)

    @staticmethod
    def load_src(file_or_src: str) -> List[dict]:
        """Load the information for the representation.

        We support file from local file/HTTP/HDFS.

        Args:
            file_or_src(`str`):
                The source YAML file or the URL points to the source file or a str
                loaded from source file.

        returns:
            The YAML file loaded as list.
        """
        # If `file_or_src` is a loacl file
        if os.path.isfile(file_or_src):
            return BaseRepr.load_file(file_or_src)
        # If `file_or_src` from HTTP
        elif file_or_src.lower().startswith('http'):
            return BaseRepr.load_url(file_or_src)
        # If `file_or_src` is neither a file nor url
        return BaseRepr.load_str(file_or_src)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.dag.base_repr import BaseRepr


class VariableRepr(BaseRepr):
    """The representation of a variable at compile-phase.

    Args:
        name:
            The representation name.
        vtype:
            This can be one of many possible variable types, such as a numpy array or
            PyTorch tensor.
        dtype:
            A string or instance of `numpy.dtype` indicating the internal data type for
            this variable.
    """
    def __init__(self, name: str, vtype: str, identifier: str = None):
        super().__init__(name)
        self._vtype = vtype
        self._id = identifier

    @property
    def vtype(self):
        return self._vtype

    # @property
    # def identifier(self):
    #     return self._id



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from towhee.cnn_trainer.training_args import TrainingArguments


class TrainerArgsTest(unittest.TestCase):
    def setUp(self) -> None:
        self.training_args = TrainingArguments(
            output_dir='./ResNet50',
            overwrite_output_dir=True,
            num_train_epochs=5,
            per_gpu_train_batch_size=4,
            prediction_loss_only=True,
        )

    def test_train_batch_size(self) -> None:
        self.assertEqual(4, self.training_args.train_batch_size)


if __name__ == '__main__':
    unittest.main(verbosity=1)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from graph_repr import GraphRepr
from operator_repr import OperatorRepr
from operator import Operator
from variable_repr import VariableReprSet


class GraphBuilder:
    """
    The IntermediateGraph builder.

    The graph construction process is driven by OperatorMetaclass and two decorators,
    @create_op_in_pipeline and @create_pipeline.
    """

    START_OP = '_start'
    END_OP = '_end'

    def __init__(self, builder_id: str = None, pipeline_func=None):
        self.builder_id = builder_id
        self._pipeline_func = pipeline_func
        self._op_index = 0
        self.graph_repr = GraphRepr(builder_id)

        # We add a pseudo start operator and a pseudo end operator to the graph.
        self._start_op = OperatorRepr(GraphBuilder.START_OP)
        self._end_op = OperatorRepr(GraphBuilder.END_OP)

    @property
    def pipeline_inputs(self):
        return self._start_op.inputs

    @pipeline_inputs.setter
    def pipeline_inputs(self, inputs: VariableReprSet):
        self._start_op.inputs = inputs
        self._start_op.outputs = self._start_op.inputs

    @property
    def pipeline_outputs(self):
        return self._end_op.outputs

    @pipeline_outputs.setter
    def pipeline_outputs(self, outputs: VariableReprSet):
        self._end_op.inputs = outputs
        self._end_op.outputs = self._end_op.inputs

    def add_op(self, name: str, op: Operator) -> OperatorRepr:
        """
        Add an OperatorRepr to the graph
        """
        # op_repr = OperatorRepr(self._generate_op_name(op), op)
        # todo: setup the op_repr
        raise NotImplementedError

    def build(self) -> GraphRepr:
        """
        Build the whole graph
        """
        raise NotImplementedError

    @property
    def _start_op(self):
        """
        The start operator is a pseudo operator. The main functionality of this
        operator is to handle the inputs of the pipeline. The start operator has no
        inputs, and its outputs are identical to the pipeline's inputs.
        """
        return self.graph_repr.op[GraphBuilder.START_OP]

    @_start_op.setter
    def _start_op(self, value: OperatorRepr):
        self.graph_repr.op[GraphBuilder.START_OP] = value

    @property
    def _end_op(self):
        """
        The start operator is a pseudo operator. The main functionality of this
        operator is to handle the inputs of the pipeline. The start operator has no
        inputs, and its outputs are identical to the pipeline's inputs.
        """
        return self.graph_repr.op[GraphBuilder.END_OP]

    @_end_op.setter
    def _end_op(self, value: OperatorRepr):
        """
        The end operator is a pseudo operator. The main functionality of this
        operator is to handle the outputs of the pipeline. The end operator has no
        outputs, and its inputs are identical to the pipeline's outputs.
        """
        self.graph_repr.op[GraphBuilder.END_OP] = value


def get_graph_builder(key: str = None) -> GraphBuilder:
    """
    Utility for getting a graph builder

    Return: if *key* is None, then try to return the graph builder of the current
        context, else return the builder related to the given *key*.
    """
    raise NotImplementedError


def delete_graph_builder(key: str = None):
    """
    Utility for deleting a graph builder
    """
    raise NotImplementedError



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import queue
import time

from towhee.engine.singleton import singleton
from towhee.tests.test_util.dataframe_test_util import MultiThreadRunner


@singleton
class Singleton:
    def __init__(self):
        pass


def create_obj(q):
    for _ in range(10):
        q.put(Singleton())
        time.sleep(0.01)


class TestSingleton(unittest.TestCase):
    """
    Singleton decorator
    """

    def test_singleton(self):
        q = queue.Queue()
        runner = MultiThreadRunner(create_obj, (q, ), 10)
        runner.start()
        runner.join()

        self.assertEqual(q.qsize(), 100)

        while not q.empty():
            self.assertEqual(q.get(), Singleton())



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from PIL import Image
from torchvision import transforms
import os

from towhee.tests.mock_operators import PYTORCH_TRANSFORME_OPERATOR_PATH, load_local_operator


class PytorchTransformTest(unittest.TestCase):
    pwd = os.getcwd()
    test_dir = pwd+'/towhee/tests/models/vit/'
    img = Image.open(test_dir + 'img.jpg')
    tfms = transforms.Compose([transforms.Resize(256),
                               transforms.CenterCrop(224),
                               transforms.ToTensor(),
                               transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
                               ])
    img1 = tfms(img).unsqueeze(0)

    def test_transform_operator(self):
        trans = load_local_operator('pytorch_transform_operator', PYTORCH_TRANSFORME_OPERATOR_PATH)
        op = trans.PytorchTransformOperator(256)
        outputs = op(self.img)
        print(type(outputs.img_transformed))
        c = (self.img1.numpy() == outputs.img_transformed.numpy())
        self.assertEqual(c.all(), True)


if __name__ == '__main__':
    unittest.main()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from collections import namedtuple

from towhee.engine.task import Task


def mock_op(k1, k2):
    Point = namedtuple('Point', ['x', 'y'])
    return Point(k1, k2)


class MockOutputsWriter:
    def __init__(self):
        self.outputs = None

    def write(self, task: Task):
        self.outputs = task.outputs._asdict()


class TestTask(unittest.TestCase):
    """
    Task basic test
    """

    def test_task_basic_func(self):
        args = {'arg1': 1, 'arg2': 'test'}
        inputs = {'k1': 1, 'k2': 'v1'}
        mock_output = MockOutputsWriter()
        task = Task('mock_op', 'mock_op', args, inputs, 0)
        task.add_finish_handler(mock_output.write)
        self.assertEqual(task.op_name, 'mock_op')
        self.assertEqual(task.inputs, inputs)
        self.assertEqual(task.op_args, args)
        task.execute(mock_op)
        self.assertEqual(task.outputs.x, inputs['k1'])
        self.assertEqual(task.outputs.y, inputs['k2'])
        self.assertEqual(mock_output.outputs, task.outputs._asdict())
        self.assertGreater(task.runtime, 0)



# coding=utf-8
# Copyright 2020-present the HuggingFace Inc. team and 2021 Zilliz.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import torchvision

from torch import optim
from towhee.cnn_trainer.trainer_callback import DefaultFlowCallback, CallbackHandler, ProgressCallback

DEFAULT_CALLBACKS = [DefaultFlowCallback]


class TrainerCallbackTest(unittest.TestCase):
    def setUp(self) -> None:
        self.model = torchvision.models.resnet50(pretrained=True)
        self.optimizer = optim.SGD(self.model.parameters(), lr=0.01, momentum=0.9)
        self.lr_scheduler = optim.lr_scheduler.StepLR(self.optimizer, step_size=3, gamma=0.1)

    def check_callbacks_equality(self, cbs1, cbs2):
        self.assertEqual(len(cbs1), len(cbs2))

        # Order doesn't matter
        cbs1 = list(sorted(cbs1, key=lambda cb: cb.__name__ if isinstance(cb, type) else cb.__class__.__name__))
        cbs2 = list(sorted(cbs2, key=lambda cb: cb.__name__ if isinstance(cb, type) else cb.__class__.__name__))
        for cb1, cb2 in zip(cbs1, cbs2):
            if isinstance(cb1, type) and isinstance(cb2, type):
                self.assertEqual(cb1, cb2)
            elif isinstance(cb1, type) and not isinstance(cb2, type):
                self.assertEqual(cb1, cb2.__class__)
            elif not isinstance(cb1, type) and isinstance(cb2, type):
                self.assertEqual(cb1.__class__, cb2)
            else:
                self.assertEqual(cb1, cb2)

    def test_add_callback(self) -> None:
        callbacks = DEFAULT_CALLBACKS.copy() + [ProgressCallback]
        callback_handler = CallbackHandler(
            callbacks, self.model, self.optimizer, self.lr_scheduler
        )
        expected_callbacks = DEFAULT_CALLBACKS.copy() + [ProgressCallback]
        self.check_callbacks_equality(callback_handler.callbacks, expected_callbacks)


if __name__ == '__main__':
    unittest.main(verbosity=1)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List, Set

from towhee.dag.base_repr import BaseRepr
from towhee.dag.dataframe_repr import DataFrameRepr
from towhee.dag.operator_repr import OperatorRepr


class GraphRepr(BaseRepr):
    """This class presents a complete representation of a graph.

    A graph contains individual subcomponents, including Operators, Dataframes, and
    Variables. Graph representations are used during execution to load functions and
    pass data to the correct operators.

    Args:
        name(`str`):
            The representation name.
        file_or_url(`str`):
            The file or remote url that stores the information of this representation.
    """

    def __init__(self, name: str, op_reprs: Dict[str, OperatorRepr], df_reprs: Dict[str, DataFrameRepr]):
        super().__init__(name)
        self._operators = op_reprs
        self._dataframes = df_reprs

    @property
    def operators(self) -> Dict[str, OperatorRepr]:
        return self._operators

    @property
    def dataframes(self) -> Dict[str, DataFrameRepr]:
        return self._dataframes

    @staticmethod
    def dfs(cur: str, adj: Dict[str, List[str]], flag: Dict[str, int], cur_list: List[str]) -> List[str]:
        """Depth-First Search the graph
        Args:
            cur(`str`):
                The name of current dataframe.
            adj(`dict`):
                A dict store the adjacent dataframe of each dataframe.
            flag(`dict`):
                A dict store the status of the columns.
                - 0 means the dataframe has not been visted yet.
                - 1 means the dataframe has been visted in this search, which means this
                columns is part of a loop.
                - 2 means the dataframe has been searched and confirmed not to be a part
                of a loop.
            cur_list('list):
                The list of dataframe that have been visited in this search.
        Returns:
            Return `False` if there is no loop, else `True` and the loop.
        """
        # if `cur` is not the input of any operator, it will not form a loop anyway
        if cur not in flag:
            return False, []
        # If `cur` has been searched and not in a loop
        if flag[cur] == 2:
            return False, []
        # If `cur` has been visited in this search
        if flag[cur] == 1:
            return True, cur_list
        flag[cur] = 1
        # Recursion
        for col in adj[cur]:
            cur_list.append(cur)
            status, loop_list = GraphRepr.dfs(col, adj, flag, cur_list)
            if status:
                return status, loop_list
        flag[cur] = 2

        return False, []

    def get_loop(self) -> List[str]:
        """Get the loop(s) inside the graph.

        Returns:
            Return the loop if exists, else an empty list.
        """
        adj = {}
        flag = {}
        dataframes = []
        # Collect adjacent information and mark all the flag as 0
        for op in self._operators.values():
            out = op.outputs[0]['df']
            for df in op.inputs:
                name = df['df']
                if name not in dataframes:
                    adj[name] = []
                    flag[name] = 0
                    dataframes.append(name)
                adj[name].append(out)
        # Recursive DFS
        for df in dataframes:
            status, loop_list = GraphRepr.dfs(df, adj, flag, [])
            if status:
                return loop_list

        return []

    def get_isolated_df(self) -> Set[str]:
        """Get the isolated dataframe(s) in the DAG.

        Returns:
            Return the isolated daatframe set if exists, else an empty set.
        """
        # First mark all the dataframes as isolated
        iso_df = {df.name for df in self._dataframes.values()}
        # If the dataframe is the input/output of any operators, remove it from isolated dataframes set
        for op in self._operators.values():
            for i in op.inputs:
                iso_df.discard(i['df'])
            iso_df.discard(op.outputs[0]['df'])

        return iso_df

    def get_isolated_op(self) -> Set[str]:
        """Get the isolated operator(s) in the DAG.

        Returns:
            Return the isolated operator set if exists, else an empty set.
        """
        # If the graph has only one operator, there is no isolated operator
        if len(self._operators.values()) == 1:
            return False, ''

        in_df = set()
        out_df = set()
        iso_op = set()
        # Collect all the input/output dataframes
        for op in self._operators.values():
            for i in op.inputs:
                in_df.add(i['df'])
            out_df.add(op.outputs[0]['df'])
        # Traverse the operators to find isolated operators
        for op in self._operators.values():
            cur_in = {i['df'] for i in op.inputs}
            cur_out = {i['df'] for i in op.outputs}
            # If the input/output are not the output/input of any other operators, the operator is isolated
            if not cur_in.intersection(out_df - cur_out) and not cur_out.intersection(in_df - cur_in):
                iso_op.add(op.name)

        return iso_op

    @staticmethod
    def from_dict(info: Dict) -> 'GraphRepr':
        # Basic schema check
        if not BaseRepr.is_valid(info, {'name', 'operators', 'dataframes'}):
            raise ValueError(
                'file or src is not a valid YAML file to describe a DAG in Towhee.'
            )
        dataframes = dict((df_info['name'], DataFrameRepr.from_dict(
            df_info)) for df_info in info['dataframes'])
        operators = dict((op_info['name'], OperatorRepr.from_dict(
            op_info)) for op_info in info['operators'])
        return GraphRepr(info['name'], operators, dataframes)

    @staticmethod
    def from_yaml(src: str):
        """Import a YAML file describing this graph.

        Args:
            src(`str`):
                YAML file (could be pre-loaded as string) to import.

        example:

            name: 'test_graph'
            operators:
                -
                    name: 'test_op_1'
                    function: 'test_function'
                    inputs:
                        -
                            df: 'test_df_1'
                            col: 0
                    outputs:
                        -
                            df: 'test_df_2'
                            col: 0
                    iter_info:
                        type: map
            dataframes:
                -
                    name: 'test_df_1'
                    columns:
                        -
                            vtype: 'int'
                    name: 'test_df_2'
                    columns:
                        -
                           vtype: 'int'
        """
        info = BaseRepr.load_src(src)
        return GraphRepr.from_dict(info)
        # Load the information and get isolation and loop
        # info = BaseRepr.load_src(src)
        # graph = GraphRepr.from_dict(info)
        # iso_df = graph.get_isolated_df()
        # iso_op = graph.get_isolated_op()
        # loop = graph.get_loop()
        # # If no loop or isolation found, return the GraphRepr
        # if not iso_df and not iso_op and not loop:
        #     return GraphRepr.from_dict(info)
        # # generate error message
        # df_msg = '' if not iso_df else f'The DAG contains isolated dataframe(s) {iso_df}.'
        # op_msg = '' if not iso_op else f'The DAG contains isolated operator(s) {iso_op}.'
        # loop_msg = '' if not loop else f'The DAG contains loop consists of {loop}'
        # msg = '.'.join([df_msg, op_msg, loop_msg]).split(None)

        # raise ValueError(msg)

    def to_yaml(self) -> str:
        """Export a YAML file describing this graph.

        Returns:
            A string with the graph's serialized contents.
        """
        # TODO(Chiiizzzy)
        raise NotImplementedError



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from setuptools import setup, find_packages
from setuptools.command.install import install
import unittest
from pathlib import Path
import shutil


def test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover('towhee/tests', pattern='test_*.py')
    return test_suite


def create_towhee_cache(src: str, dst: str, is_dir: bool = False):
    if not Path(dst).is_dir():
        Path(dst).mkdir(parents=True)
        if is_dir:
            dst = os.path.join(dst, src.split('/')[-1])
            shutil.copytree(src, dst)
        else:
            shutil.copy(src, dst)


class PostInstallCommand(install):
    def run(self):
        from towhee.engine import LOCAL_OPERATOR_CACHE, LOCAL_PIPELINE_CACHE, MOCK_OPS, MOCK_PIPELINES
        install.run(self)
        create_towhee_cache(MOCK_PIPELINES, LOCAL_PIPELINE_CACHE)
        create_towhee_cache(MOCK_OPS, LOCAL_OPERATOR_CACHE, True)


setup(name="towhee_test",
      version="0.1.0",
      description="",
      author="Towhee Team",
      author_email="towhee-team@zilliz.com",
      url="https://github.com/towhee-io/towhee",
      test_suite="setup.test_suite",
      cmdclass={'install': PostInstallCommand},
      install_requires=[
          'torch>=1.2.0',
          'torchvision>=0.4.0',
          'numpy>=1.20.3',
          'pandas>=1.2.4',
          'pyyaml>=5.3',
          'requests>=2.25',
          'tqdm>=4.59.0',
          'pillow>=8.3.1',
          'scipy>=1.7.1',
      ],
      packages=find_packages(),
      package_data={'towhee.tests.test_util': ['*.yaml']},
      license="http://www.apache.org/licenses/LICENSE-2.0")



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class HandlerMixin:
    """
    This mixin creates handler management interfaces
    Args:
        handler_prefix: (`str`)
            `HandlerMixin` will create a set of handler management interfaces based on
            each handler_prefix.
            For instance, given a `handler_prefix`='task_finish', there will be two
            methods that are generated, `add_task_finish_handler`
            and `call_task_finish_handlers`.
    """

    def add_handler_methods(self, *handler_prefix):

        class _HandlerMgr:
            """
            Each type of handlers are managed by one `_HandlerMgr`
            """

            def __init__(self):
                self._handlers = []

            def add_handler(self, handler):
                if isinstance(handler, list):
                    for h in handler:
                        if not callable(h):
                            raise AttributeError('handler must be callable')
                    self._handlers += handler
                elif callable(handler):
                    self._handlers.append(handler)
                else:
                    raise AttributeError('handler must be callable')

            def call_handlers(self, *args):
                for h in self._handlers:
                    h(*args)

        self._handler_mgrs = []

        for prefix in handler_prefix:
            if isinstance(prefix, str):
                handler_mgr = _HandlerMgr()
                self._handler_mgrs.append(handler_mgr)

                add_handler_method_name = 'add_' + prefix + '_handler'
                self.__setattr__(add_handler_method_name, handler_mgr.add_handler)
                call_handler_method_name = 'call_' + prefix + '_handlers'
                self.__setattr__(call_handler_method_name, handler_mgr.call_handlers)
            else:
                raise AttributeError(
                    '`HandlerMixin` can only take `str` as handler prefix, but found %s.'
                    % (type(prefix)))



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


#import unittest


#from towhee.engine._repr_to_ctx import create_ctxs


src = """
name: 'test'
 operators:
    -
        name: 'test_op_1'
        function: 'test_function'
        init_args:
            arg1: 1
            arg2: 'test'
        inputs:
            -
                name: 'k1'
                df: 'test_df_1'
                col: 0
        outputs:
            -
                df: 'test_df_2'
        iter_info:
            type: map
    -
        name: 'test_op_2'
        function: 'test_function'
        init_args:
            arg1: 1
            arg2: 'test'
        inputs:
            -
                name: 'k1'
                df: 'test_df_2'
                col: 2
        outputs:
            -
                df: 'test_df_3'
        iter_info:
            type: map
 dataframes:
    -
        name: 'test_df_1'
        columns:
            -
                name: 'k1'
                vtype: 'int'
            -
                name: 'k2'
                vtype: 'int'
    -
        name: 'test_df_2'
        columns:
            -
                name: 'k1'
                vtype: 'int'
    -
        name: 'test_df_3'
        columns:
            -
                name: 'k1'
                vtype: 'int'
 """


# class TestReprToCtxs(unittest.TestCase):
#    """
#    Test repr to ctxs
#    """
#
#    def test_to_ctxs(self):
#        g_ctx, dataframes = create_ctxs(src)
#        self.assertEqual(len(g_ctx.op_ctxs), 2)
#        self.assertEqual(len(dataframes), 3)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from towhee.operator import Operator
from towhee.engine.operator_loader import OperatorLoader
from towhee.engine.task import Task


class OperatorPool:
    """`OperatorPool` manages `Operator` creation, acquisition, release, and garbage
    collection. Each `TaskExecutor` has one `OperatorPool`.
    """

    def __init__(self, cache_path: str = None):
        self._op_loader = OperatorLoader(cache_path)
        self._all_ops = {}

    @property
    def available_ops(self):
        return self._all_ops.keys()

    def is_op_available(self, task: Task) -> bool:
        """Determines whether an operator that can be used to fulfill the given Task is
        currently loaded.

        Args:
            task: (`towhee.Task`)
                Task

        Returns:
            (`bool`)
                Returns `True` if the specified input task can be run without having to
                load a new operator from cache.
        """
        return task.op_key in self._all_ops

    def acquire_op(self, task: Task) -> Operator:
        """Given a `Task`, instruct the `OperatorPool` to reserve and return the
        specified operator for use in the executor.

        Args:
            task: (`towhee.Task`)
                Task to acquire an operator for.

        Returns:
            (`towhee.operator.Operator`)
                The operator instance reserved for the caller.
        """

        # Load the operator if the computed key does not exist in the operator
        # dictionary.
        op_key = task.op_key
        if op_key not in self._all_ops:
            op = self._op_loader.load_operator(task.hub_op_id, task.op_args)
            op.key = op_key
        else:
            op = self._all_ops[op_key]

        # Let there be a record of the operator existing in the pool, but remove its
        # pointer until the operator is released by the `TaskExecutor`.
        self._all_ops[op_key] = None

        return op

    def release_op(self, op: Operator):
        """Releases the specified operator and all associated resources back to the
        `OperatorPool`.

        Args:
            op: (`towhee.Operator`)
                `Operator` instance to add back into the operator pool.
        """
        self._all_ops[op.key] = op



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
from typing import Any, Tuple
import os

from towhee.dataframe import DataFrame
from towhee.dataframe import Variable
from towhee.engine.engine import Engine, EngineConfig, start_engine
from towhee.engine.pipeline import Pipeline
from towhee.engine import LOCAL_PIPELINE_CACHE


__all__ = [
    'DEFAULT_PIPELINES',
    'pipeline'
]


DEFAULT_PIPELINES = {
    'image-embedding': 'resnet50_embedding'
}

_PIPELINE_CACHE_ENV = 'PIPELINE_CACHE'


class _PipelineWrapper:
    """A wrapper class around `Pipeline` which prevents users from having to create
    `DataFrame` instances by hand.

    Args:
        pipeline: `towhee.Pipeline`
            Base `Pipeline` instance for which this object will provide a wrapper for.
    """

    def __init__(self, pipeline_: Pipeline):
        self._pipeline = pipeline_

    def __call__(self, *args) -> Tuple[Any]:
        """Wraps the input arguments around a `Dataframe` for Pipeline.__call__().
        """

        # Create `Variable` tuple from input arguments.
        vargs = []
        for arg in args:
            vtype = type(arg).__name__
            vargs.append(Variable(vtype, arg))
        vargs = tuple(vargs)

        # Process the data through the pipeline.
        in_df = DataFrame('_in_df')
        in_df.put(vargs)
        out_df = self._pipeline(in_df)

        # Extract values from output tuple
        res = []
        for v in out_df.get(0, out_df.size)[1][0]:
            res.append(v.value)

        return tuple(res)


def _get_pipeline_cache(cache_path: str):
    if not cache_path:
        cache_path = os.environ.get(_PIPELINE_CACHE_ENV) if os.environ.get(
            _PIPELINE_CACHE_ENV) else LOCAL_PIPELINE_CACHE
    return Path(cache_path)


# def _get_hello_towhee_pipeline():
#     return Path(__file__).parent / 'tests/test_util/resnet50_embedding.yaml'


def pipeline(task: str, cache: str = None):
    """Entry method which takes either an input task or path to an operator YAML.

    Args:
        task: (`str`)
            Task name or YAML file location to use.
    """

    # If the task name coincides with one of the default pipelines, use the YAML
    # specified by that default pipeline instead of trying to lookup a YAML in the hub
    # or cache.

    start_engine()
    # TODO (jiangjunjie) delete when hub is ready
    # if task.startswith('hello_towhee'):
    #     yaml_path = _get_hello_towhee_pipeline()
    # else:
    task = DEFAULT_PIPELINES.get(task, task)

    # Get YAML path given task name. The default cache location for pipelines is
    # $HOME/.towhee/pipelines
    # TODO(fzliu): if pipeline is not available in cache, acquire it from hub
    cache_path = _get_pipeline_cache(cache)
    yaml_path = cache_path / (task + '.yaml')

    if not yaml_path.is_file():
        raise NameError('Can not find pipeline by name %s ' % task)

    # Create `Pipeline` object given its graph representation, then add the pipeline to
    # the existing `Engine`.
    engine = Engine()
    pipeline_ = Pipeline(str(yaml_path))
    engine.add_pipeline(pipeline_)

    return _PipelineWrapper(pipeline_)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from towhee.dag.graph_repr import GraphRepr
from towhee.dag.variable_repr import VariableRepr
from towhee.dag.dataframe_repr import DataFrameRepr
import unittest

from towhee.engine.engine import Engine, EngineConfig
from towhee.tests import CACHE_PATH
from towhee.engine.pipeline import Pipeline
from towhee.dataframe import DataFrame, Variable
from towhee.tests.test_util import SIMPLE_PIPELINE_YAML
from towhee.dag import OperatorRepr


class TestEngine(unittest.TestCase):
    """
      combine tests of engine/scheduler/task-executor/task
      """

    def setUp(self):
        conf = EngineConfig()
        conf.cache_path = CACHE_PATH
        conf.sched_interval_ms = 20
        engine = Engine()
        if not engine.is_alive():
            engine.start()

    def test_engine(self):

        start_op_repr = OperatorRepr(
            name='_start_op',
            function='internal',
            init_args={},
            inputs=[
                {'name': 'num', 'df': 'op_test_in', 'col': 0}
            ],
            outputs=[{'df': 'op_test_in'}],
            iter_info={'type': 'map'}
        )

        add_op_repr = OperatorRepr(
            name='mock_operators/add_operator',
            function='mock_operators/add_operator',
            init_args={'factor': 2},
            inputs=[
                {'name': 'num', 'df': 'op_test_in', 'col': 0}
            ],
            outputs=[{'df': 'op_test_out'}],
            iter_info={'type': 'map'}
        )

        end_op_repr = OperatorRepr(
            name='_end_op',
            function='internal',
            init_args={},
            inputs=[
                {'name': 'sum', 'df': 'op_test_out', 'col': 0}
            ],
            outputs=[{'df': 'op_test_out'}],
            iter_info={'type': 'map'}
        )

        df_in_repr = DataFrameRepr(
            name='op_test_in',
            columns=[VariableRepr('num', 'int')]
        )

        df_out_repr = DataFrameRepr(
            name='op_test_out',
            columns=[VariableRepr('sum', 'int')]
        )

        op_reprs = {
            start_op_repr.name: start_op_repr,
            add_op_repr.name: add_op_repr,
            end_op_repr.name: end_op_repr
        }

        df_reprs = {
            df_in_repr.name: df_in_repr,
            df_out_repr.name: df_out_repr
        }

        graph_repr = GraphRepr('add', op_reprs, df_reprs)

        df_in = DataFrame(
            'op_test_in', {'sum': {'index': 0, 'type': 'int'}})
        df_in.put((Variable('int', 1), ))
        df_in.seal()

        self._pipeline = Pipeline(graph_repr)
        engine = Engine()
        engine.add_pipeline(self._pipeline)
        result = self._pipeline(df_in)
        _, ret = result.get(0, 1)
        self.assertEqual(ret[0][0].value, 3)

        df_in = DataFrame(
            'op_test_in', {'sum': {'index': 0, 'type': 'int'}})
        df_in.put((Variable('int', 3), ))
        df_in.seal()

        result = self._pipeline(df_in)
        _, ret = result.get(0, 1)
        self.assertEqual(ret[0][0].value, 5)

    def test_simple_pipeline(self):
        with open(SIMPLE_PIPELINE_YAML, 'r', encoding='utf-8') as f:
            p = Pipeline(f.read())
        engine = Engine()
        engine.add_pipeline(p)

        df_in = DataFrame(
            'inputs', {'sum': {'index': 0, 'type': 'int'}})

        df_in.put((Variable('int', 3), ))
        df_in.seal()
        result = p(df_in)
        _, ret = result.get(0, 1)
        self.assertEqual(ret[0][0].value, 6)

        df_in = DataFrame(
            'inputs', {'sum': {'index': 0, 'type': 'int'}})

        df_in.put((Variable('int', 7), ))
        df_in.seal()
        result = p(df_in)
        _, ret = result.get(0, 1)
        self.assertEqual(ret[0][0].value, 10)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NamedTuple
import torch
from torchvision import transforms

from towhee.operator import Operator


class PytorchTransformOperator(Operator):
    """
    Image transform operator
        Args:
            tfms(transforms.Compose):
                This is used to transform an image.
    """

    def __init__(self, size: int) -> None:
        super().__init__()
        # user defined transform
        self.tfms = transforms.Compose([transforms.Resize(size),
                                        transforms.CenterCrop(224),
                                        transforms.ToTensor(),
                                        transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
                                        ])

    def __call__(self, img_tensor: torch.Tensor) -> NamedTuple('Outputs', [('img_transformed', torch.Tensor)]):
        Outputs = NamedTuple('Outputs', [('img_transformed', torch.Tensor)])
        return Outputs(self.tfms(img_tensor).unsqueeze(0))



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.dag.graph_repr import GraphRepr

from towhee.engine.graph_context import GraphContext
from towhee.dataframe import DataFrame


class Pipeline:
    """
    The runtime pipeline context, include graph context, all dataframes
    """

    def __init__(self, graph_repr: GraphRepr, parallelism: int = 1) -> None:
        """
        Args:
            graph_repr: (yaml `str` or `towhee.dag.GraphRepr`)
                The graph representation
            parallelism: (`int`)
                how many rows of inputs to be processed concurrently
        """
        if isinstance(graph_repr, str):
            self._graph_repr = GraphRepr.from_yaml(graph_repr)
        else:
            self._graph_repr = graph_repr

        self._parallelism = parallelism
        self.on_graph_finish_handlers = []
        self._scheduler = None

        # self._cv = threading.Condition()
        # self.on_graph_finish_handlers = [self._on_graph_finish]
        # self._build()

    def register(self, scheduler):
        self._scheduler = scheduler

    def __call__(self, inputs: DataFrame) -> DataFrame:
        """
        Process one input data
        """
        if self._scheduler is None:
            raise AttributeError(
                "The pipeline is not registered to a scheduler")

        assert inputs.size == 1
        g = GraphContext(0, self._graph_repr)
        self._scheduler.register(g)
        _, data = inputs.get(0, 1)
        g(data[0])
        return g.result()

    # @property
    # def graph_contexts(self):
    #     return self._graph_ctxs

    # @property
    # def is_busy(self) -> bool:
    #     return False not in [graph_ctx.is_busy for graph_ctx in self._graph_ctxs]

    # def __call__(self, inputs: DataFrame) -> DataFrame:

    #     def _feed_one_graph_ctx(row):
    #         for graph_ctx in self._graph_ctxs:
    #             if not graph_ctx.is_busy:
    #                 # fill the input row to the available graph context
    #                 graph_ctx(row)
    #                 return

    #     self.outputs = DataFrame()

    #     for row in inputs.map_iter():
    #         if not self.is_busy:
    #             _feed_one_graph_ctx(row)
    #         else:
    #             with self._cv:
    #                 if not self.is_busy:
    #                     _feed_one_graph_ctx(row)
    #                 else:
    #                     self._cv.wait()

    #     # todo: GuoRentong, need to track each graph call for multi-row inputs
    #     with self._cv:
    #         if self.is_busy:
    #             self._cv.wait()
    #         return self.outputs

    # def _on_graph_finish(self, graph_ctx: GraphContext):
    #     # Organize the GraphContext's output into Pipeline's outputs.
    #     self.outputs.merge(graph_ctx.outputs)
    #     # Notify the run loop that a GraphContext is in idle state.
    #     with self._cv:
    #         self._cv.notify()

    # def _build(self):
    #     """
    #     Create GraphContexts and set up input iterators.
    #     """
    #     build graph contexts
    #     self._graph_ctxs = [GraphContext(i, self._graph_repr)
    #                         for i in range(self._parallelism)]

    #     # add on_task_finish_handlers to graph contexts
    #     for g in self._graph_ctxs:
    #         g.on_finish_handlers += self.on_graph_finish_handlers



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import signal
import threading
import logging
import time

from towhee.engine.operator_pool import OperatorPool
from towhee.engine.task import Task
from towhee.engine.task_queue import TaskQueue


class TaskExecutor(threading.Thread):
    """A FIFO task executor. Each device has one TaskExecutor.

    Args:
        name: (`str`)
            Name of the device for which the executor will run. Example device name:
                dev_name_0 = 'cpu:0'
                dev_name_1 = 'gpu:0'
        cache_path: (`str`)
            Local path for which operators are stored. Defaults to
            `$HOME/.towhee/operators`.
    """

    def __init__(self, name: str, cache_path: str = None):
        super().__init__()
        self._name = name
        self._task_queue = TaskQueue()
        self._op_pool = OperatorPool(cache_path=cache_path)
        self._is_run = True
        self.setDaemon(True)

    @property
    def name(self):
        return self._name

    @property
    def num_tasks(self):
        return self._task_queue.size

    def is_op_available(self, task: Task) -> bool:
        return self._op_pool.is_op_availble(task)

    # def set_op_parallelism(self, name: str, parallel: int = 1):
    #     """
    #     Set how many Operators with same name can be called concurrently.
    #     """
    #     raise NotImplementedError

    def push_task(self, task: Task) -> bool:
        """Push a task to the end of the task queue.

        Args:
            task:
                Pre-initialized `Task` object to push onto the queue.
        """
        return self._task_queue.push(task)

    def run(self):
        """Runs the execution loop.
        """

        # Continue execution until the run flag is externally set to false and the queue
        # is entirely empty of tasks.
        # TODO(fzliu): use ThreadPool or manually create multiple threads to handle
        # tasks which are I/O bound.
        while self._is_run or not self._task_queue.empty:

            # If there are no tasks in the queue, this might return `None` after
            # blocking for a while.
            task = self._task_queue.pop()

            # Execute tasks in first-in, first-out fashion. The `Scheduler` is
            # ultimately the one responsible for determining which tasks get executed
            # before others.
            if task:
                try:
                    op = self._op_pool.acquire_op(task)
                except Exception as e:  # pylint: disable=broad-except
                    logging.error(e)

                    # TODO (junjie.jiangjjj) gracefull exit
                    os.kill(os.getpid(), signal.SIGINT)

                task.execute(op)
                self._op_pool.release_op(op)

            # TODO(fzliu): Improve wait logic if an empty queue is detected.
            else:
                time.sleep(0.01)

    def stop(self):
        """Sets a flag, which stops the execution loop after a period of time.
        """
        self._is_run = False
        self.join()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import NamedTuple

import numpy
import torch
import torchvision

from towhee.operator import Operator


class PytorchCnnOperator(Operator):
    """
    PyTorch model operator base
    """
    def __init__(self, model_name) -> None:
        super().__init__()
        model_func = getattr(torchvision.models, model_name)
        self._model = model_func(pretrained=True)
        self._model.eval()

    def __call__(self, img_tensor: torch.Tensor) -> NamedTuple('Outputs', [('cnn', numpy.ndarray)]):
        Outputs = NamedTuple('Outputs', [('cnn', numpy.ndarray)])
        return Outputs(self._model(img_tensor).detach().numpy())

    def train(self):
        """
        For training model
        """
        pass
